/*
 * Copyright (c) 2021, 2023, 2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jul 27 2021
 */

#include "acgazeboft.h"

#include <err.h>
#include <fcntl.h>
#include <stdlib.h>

#include <vector>

#include "gazeboft_c_types.h"
#include "codels.h"


/* --- Function get_sensors --------------------------------------------- */

/** Codel gzft_get_sensors of function get_sensors.
 *
 * Returns genom_ok.
 * Throws gazeboft_e_nodev.
 */
genom_event
gzft_get_sensors(sequence16_string256 *sensors,
                 const genom_context self)
{
  gz::transport::Node node;
  std::vector<std::string> topics;
  std::vector<gz::transport::MessagePublisher> pubs, subs;

  node.TopicList(topics);

  sensors->_length = 0;
  for (auto const &t : topics) {
    if (sensors->_length >= sensors->_maximum) break;

    if (!node.TopicInfo(t, pubs, subs)) continue;
    for (auto const &p : pubs) {
      if (p.MsgTypeName() != "gz.msgs.Wrench") break;

      snprintf(sensors->_buffer[sensors->_length],
               sizeof(*sensors->_buffer), "%s", t.c_str());
      sensors->_length++;
      break;
    }
  }

  return genom_ok;
}


/* --- Function log ----------------------------------------------------- */

/** Codel gzft_log_start of function log.
 *
 * Returns genom_ok.
 * Throws gazeboft_e_sys.
 */
genom_event
gzft_log_start(const char path[64], uint32_t decimation,
               gazeboft_log_s **log, const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return gazeboft_e_sys_error(path, self);

  if (write(fd, gazeboft_log_header "\n", sizeof(gazeboft_log_header)) < 0)
    return gazeboft_e_sys_error(path, self);

  if ((*log)->req.aio_fildes >= 0) {
    close((*log)->req.aio_fildes);

    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->req.aio_fildes = fd;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel gzft_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
gzft_log_stop(gazeboft_log_s **log, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel gzft_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
gzft_log_info(const gazeboft_log_s *log, uint32_t *miss,
              uint32_t *total, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }
  return genom_ok;
}
