/*
 * Copyright (c) 2021, 2023, 2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jul 27 2021
 */

#include "acgazeboft.h"

#include <err.h>

#include "gazeboft_c_types.h"
#include "codels.h"


/* --- Task read -------------------------------------------------------- */


/** Codel gzft_init of task read.
 *
 * Triggered by gazeboft_start.
 * Yields to gazeboft_pause_read.
 */
genom_event
gzft_init(gazeboft_ids *ids, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  /* default to no calibration */
  ids->calibration = (struct gazeboft_ids_calibration_s){
    .b = {0}
  };

  /* init data */
  ids->data = nullptr;

  /* init logging */
  ids->log = (gazeboft_log_s *)calloc(1, sizeof(*ids->log));
  if (!ids->log) return gazeboft_e_sys_error("malloc", self);

  ids->log->req.aio_fildes = -1;
  ids->log->req.aio_offset = 0;
  ids->log->req.aio_buf = ids->log->buffer;
  ids->log->req.aio_nbytes = 0;
  ids->log->req.aio_reqprio = 0;
  ids->log->req.aio_sigevent.sigev_notify = SIGEV_NONE;
  ids->log->req.aio_lio_opcode = LIO_NOP;

  ids->log->pending = false;
  ids->log->skipped = false;
  ids->log->decimation = 1;
  ids->log->missed = 0;
  ids->log->total = 0;

  return gazeboft_pause_read;
}


/** Codel gzft_read of task read.
 *
 * Triggered by gazeboft_read.
 * Yields to gazeboft_export, gazeboft_pause_read.
 */
genom_event
gzft_read(const gazeboft_topic *data, or_wrench_estimator_state *raw,
          const genom_context self)
{
  if (!data) return gazeboft_pause_read;

  std::unique_lock<std::mutex> lk(data->m);
  if (data->cv.wait_for(
        lk, std::chrono::milliseconds{500}, [data]{ return data->fresh; })) {
    data->fresh = false;
    *raw = data->w;
    return gazeboft_export;
  }

  return gazeboft_pause_read;
}


/** Codel gzft_export of task read.
 *
 * Triggered by gazeboft_export.
 * Yields to gazeboft_read.
 */
genom_event
gzft_export(const or_wrench_estimator_state *raw,
            const gazeboft_ids_calibration_s *calibration,
            gazeboft_log_s **log, const gazeboft_wrench *wrench,
            const genom_context self)
{
  or_wrench_estimator_state *wdata = wrench->data(self);

  /* apply bias */
  *wdata = *raw;
  wdata->force._value.x += calibration->b[0];
  wdata->force._value.y += calibration->b[1];
  wdata->force._value.z += calibration->b[2];
  wdata->torque._value.x += calibration->b[3];
  wdata->torque._value.y += calibration->b[4];
  wdata->torque._value.z += calibration->b[5];

  /* publish */
  wrench->write(self);

  /* log */
  if ((*log)->req.aio_fildes >= 0) {
    (*log)->total++;
    if ((*log)->total % (*log)->decimation == 0) {
      if ((*log)->pending) {
        if (aio_error(&(*log)->req) != EINPROGRESS) {
          (*log)->pending = false;
          if (aio_return(&(*log)->req) <= 0) {
            warn("log");
            close((*log)->req.aio_fildes);
            (*log)->req.aio_fildes = -1;
          }
        } else {
          (*log)->skipped = true;
          (*log)->missed++;
        }
      }
    }

    if ((*log)->req.aio_fildes >= 0 && !(*log)->pending) {
      (*log)->req.aio_nbytes = snprintf(
        (*log)->buffer, sizeof((*log)->buffer),
        "%s" gazeboft_log_line "\n",
        (*log)->skipped ? "\n" : "",
        (uint64_t)wdata->ts.sec, (uint32_t)wdata->ts.nsec,
        wdata->force._value.x, wdata->force._value.y, wdata->force._value.z,
        wdata->torque._value.x, wdata->torque._value.y, wdata->torque._value.z);

      if (aio_write(&(*log)->req)) {
        warn("log");
        close((*log)->req.aio_fildes);
        (*log)->req.aio_fildes = -1;
      } else
        (*log)->pending = true;

      (*log)->skipped = false;
    }
  }

  return gazeboft_read;
}


/** Codel gzft_stop of task read.
 *
 * Triggered by gazeboft_stop.
 * Yields to gazeboft_ether.
 */
genom_event
gzft_stop(gazeboft_log_s **log, gazeboft_topic **data,
          const genom_context self)
{
  gzft_log_stop(log, self);
  if (*log) free(*log);

  gzft_disconnect_start(data, self);

  return gazeboft_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel gzft_connect_start of activity connect.
 *
 * Triggered by gazeboft_start.
 * Yields to gazeboft_ether.
 * Throws gazeboft_e_nodev, gazeboft_e_sys.
 */
genom_event
gzft_connect_start(const char topic[256], gazeboft_topic **data,
                   const genom_context self)
{
  /* drop any pending connection and reconnect */
  gzft_disconnect_start(data, self);
  gz::transport::Node *node = new gz::transport::Node;
  if (!node) return gazeboft_e_sys_error("new", self);

  /* check in topic list */
  {
    std::vector<gz::transport::MessagePublisher> pubs, subs;
    bool found = false;

    if (node->TopicInfo(topic, pubs, subs))
      for (auto const &p : pubs) {
        if (p.MsgTypeName() == "gz.msgs.Wrench") found = true;
        break;
      }

    if (!found) return gazeboft_e_nodev(self);
  }

  /* subscribe */
  *data = new gazeboft_topic;
  if (!*data) {
    delete node;
    return gazeboft_e_sys_error("new", self);
  }
  (*data)->node = node;

  if (!node->Subscribe(topic, &gazeboft_topic::cb, *data))
    return gazeboft_e_nodev(self);

  warnx("connected to %s", topic);
  return gazeboft_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel gzft_disconnect_start of activity disconnect.
 *
 * Triggered by gazeboft_start.
 * Yields to gazeboft_ether.
 */
genom_event
gzft_disconnect_start(gazeboft_topic **data, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*data) {
    for(const auto &t : (*data)->node->SubscribedTopics())
      warnx("disconnected from %s", t.c_str());
    if ((*data)->node) delete (*data)->node;
    delete *data;
    *data = nullptr;
  }
  return gazeboft_ether;
}


/* --- Activity set_zero ------------------------------------------------ */

/** Codel gzft_zero_start of activity set_zero.
 *
 * Triggered by gazeboft_start.
 * Yields to gazeboft_loop.
 * Throws gazeboft_e_nodev.
 */
genom_event
gzft_zero_start(const gazeboft_topic *data, int32_t *n, or_time_ts *ts,
                double ib[6], const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  int i;

  if (!data) return gazeboft_e_nodev(self);

  *n = 0;
  *ts = { 0, 0 };
  for(i=0; i<6; i++) ib[i] = 0.;
  return gazeboft_loop;
}

/** Codel gzft_zero_loop of activity set_zero.
 *
 * Triggered by gazeboft_loop.
 * Yields to gazeboft_pause_loop, gazeboft_set.
 * Throws gazeboft_e_nodev.
 */
genom_event
gzft_zero_loop(const or_wrench_estimator_state *raw, int32_t samples,
               double ib[6], or_time_ts *ts, int32_t *n,
               const genom_context self)
{
  /* this is called twice per sample, because of the read + export codels */
  if (raw->ts.sec == ts->sec && raw->ts.nsec == ts->nsec)
    return gazeboft_pause_loop;
  if ((*n)++ >= samples) return gazeboft_set;
  *ts = raw->ts;

  ib[0] -= raw->force._value.x / samples;
  ib[1] -= raw->force._value.y / samples;
  ib[2] -= raw->force._value.z / samples;
  ib[3] -= raw->torque._value.x / samples;
  ib[4] -= raw->torque._value.y / samples;
  ib[5] -= raw->torque._value.z / samples;

  return gazeboft_pause_loop;
}

/** Codel gzft_zero_set of activity set_zero.
 *
 * Triggered by gazeboft_set.
 * Yields to gazeboft_ether.
 * Throws gazeboft_e_nodev.
 */
genom_event
gzft_zero_set(const double ib[6], double b[6],
              const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  int i;

  for(i = 0; i < 6; i++)
    b[i] = ib[i];

  return gazeboft_ether;
}


/* --- gazeboft_topic::cb -------------------------------------------------- */

void
gazeboft_topic::cb(const gz::msgs::Wrench &msg)
{
  struct timeval tv;

  std::lock_guard<std::mutex> lk(m);

  gettimeofday(&tv, NULL);
  w.ts.sec = tv.tv_sec;
  w.ts.nsec = 1000 * tv.tv_usec;
  w.intrinsic = true;

  w.force._present = true;
  w.force._value.x = msg.force().x();
  w.force._value.y = msg.force().y();
  w.force._value.z = msg.force().z();
  w.force_cov._present = false;

  w.torque._present = true;
  w.torque._value.x = msg.torque().x();
  w.torque._value.y = msg.torque().y();
  w.torque._value.z = msg.torque().z();
  w.torque_cov._present = false;

  fresh = true;
  cv.notify_one();
}
